#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Package to use a "Phenology Available" image collection in Google Earth Engine."""

__all__ = (
    "__title__", "__summary__", "__uri__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
)

__title__ = "GEE phenology"
__summary__ = "Generate a 'Phenology Available Result' image in Google Earth Engine"
__uri__ = "https://github.com/francbartoli/gee_pheno"

__author__ = """Francesco Bartoli"""
__email__ = 'francesco.bartoli@geobeyond.it'

__version__ = '0.1.5'

__license__ = "GNU GENERAL PUBLIC LICENSE, Version 3"
__copyright__ = "Francesco Bartoli"



# -*- coding: utf-8 -*-

from ee import Filter as FilterEE
from ee import Image as ImageEE
from ee import ImageCollection as ImageCollectionEE
from ee import EEException
import daiquiri
from exception import GEEException
from filters import (
    PHEStartSeason,
    PHEEndSeason,
    GrowingSeasonDate
)
from util.dekad2millis import PHEDekad2Millis


class Phenology(object):
    """ Calculate GEE phenology metrics for each record
    GEE phenology metrics describe the general season start and end
    using an algorithm by Livia Peiser from FAO, 2017 based
    on dekads.
    Attributes:
        self.pheno (xarray.Dataset): Xarray structured dataset containing
            phenology metrics. These metrics include:
            * PHEsos_img: annual time series of the start of season (ee.Image)
            * PHEeos_img: annual time series of the end of season (ee.Image)
            * ini_date: filter start date around target year (ee.Date)
            * end_date: filter end date around target year (ee.Date)
            * PHEstart: start date expressed in millis from phenology dekad
                number (ee.Image)
            * PHEend: end date expressed in millis from phenology dekad
                number (ee.Image)
            * season_s: season start expressed in millis from phenology
                dekad number (ee.Image)
            * season_e: season end expressed in millis from phenology
                dekad number (ee.Image)

    Args:
        PHE_coll (ee.ImageCollection, mandatory): multi-dimensional raster of
            Phenology data
        year (string, mandatory): target year of interest
        season (int, mandatory): season of interest

    """

    def __init__(self, **kwargs):
        logger = daiquiri.getLogger(__name__, subsystem="pheno")
        self.logger = logger
        self.logger.info("==========INIT Phenology===========")
        super(Phenology, self).__init__()
        self.logger.debug("===phe_coll====> {0}".format(
            kwargs.get("phe_coll")
        ))
        self.PHE_coll = self._phecoll(kwargs.get("phe_coll"))
        self.logger.debug("===year====> {0}".format(
            kwargs.get("year")
        ))
        self.year = kwargs.get("year")
        self.logger.debug("===season====> {0}".format(
            kwargs.get("season")
        ))
        self.season = kwargs.get("season")

    def _phecoll(self, asset_id):
        try:
            imgColl = ImageCollectionEE(asset_id)
            return imgColl
        except EEException as e:
            raise GEEException(e.message)

    @property
    def start_year(self):
        try:
            starty = self.PHE_coll.sort('system:time_start', True).first().get(
                'time_extent'
            ).getInfo()[5:9]
            return int(starty)
        except EEException as e:
            raise GEEException(e.message)

    @property
    def end_year(self):
        try:
            endy = self.PHE_coll.sort('system:time_start', False).first().get(
                'time_extent'
            ).getInfo()[5:9]
            return int(endy)
        except EEException as e:
            raise GEEException(e.message)

    @property
    def PHEsos_img(self):
        kw_soy = {
            "year": self.year,
            "season": self.season
        }
        return PHEStartSeason(**kw_soy).apply(self.PHE_coll)

    @property
    def PHEeos_img(self):
        kw_soy = {
            "year": self.year,
            "season": self.season
        }
        return PHEEndSeason(**kw_soy).apply(self.PHE_coll)

    @property
    def ini_date(self):
        kw_soy = {
            "year": self.year
        }
        return GrowingSeasonDate(**kw_soy).apply()[0]

    @property
    def end_date(self):
        kw_soy = {
            "year": self.year
        }
        return GrowingSeasonDate(**kw_soy).apply()[1]

    @property
    def PHEstart(self):
        kw_phety = {
            "phe_img": self.PHEsos_img,
            "tyear": self.year
        }
        return PHEDekad2Millis(**kw_phety).apply(
            self.start_year,
            self.end_year
        )

    @property
    def PHEend(self):
        kw_phety = {
            "phe_img": self.PHEeos_img,
            "tyear": self.year
        }
        return PHEDekad2Millis(**kw_phety).apply(
            self.start_year,
            self.end_year
        )

    @property
    def season_s(self):
        return self.PHEstart.select([4])

    @property
    def season_e(self):
        return self.PHEend.select([4])



# -*- coding: utf-8 -*-

""" Module holding custom filters for image collections """

from abc import ABCMeta, abstractmethod
from ee import Filter as FilterEE
from ee import Image as ImageEE
from ee import Date as DateEE
import daiquiri
from registry_decorator import *
from exception import GEEException

__all__ = []
factory = {}


class Filter(object):
    """ Abstract Base class for filters """
    __metaclass__ = ABCMeta
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def apply(self, coll_ee, *args, **kwargs):
        pass

@register(factory)
@register_all(__all__)
class PHESeason(Filter):
    """ Generic growing season filter

    :param season: all images within this will be filtered. Goes from 1
        to 2
    :type season: int
    :param year: all images over this will be filtered. Goes from 1970
        to now
    :type year: str
    :param kwargs:
    """

    def __init__(self, filters=None, *args, **kwargs):

        logger = daiquiri.getLogger(__name__, subsystem="filters")
        self.logger = logger

        super(PHESeason, self).__init__(filters=None, **kwargs)

        if not filters:
            self.filters = []
            self.logger.info("Filters are empty so they will be populated")
            self.logger.debug("Kwargs are: {0}".format(
                kwargs.items()
            ))
            for k, v in kwargs.items():
                logger.debug("key/value pair of filter is {0},{1}".format(
                    k, v
                ))
                # workaround because the property "year"
                # is not presente anymore
                if k == "year":
                    k = "time_extent"
                    v = "from {0}-01-01 to {0}-12-31".format(v)
                self.filters.append(
                    FilterEE.eq(k, v)
                )
        else:
            self.filters = filters

        self.logger.info("Filters are populated".format(
            self.filters
        ))
        for filter in self.filters:
            self.logger.debug("=== filter is ====> {0}".format(
                filter.getInfo()
            ))

    def apply(self, coll_ee, *args, **kwargs):
        """ Apply the filter

        :param col_ee: the image collection to apply the filter
        :type col_ee: ee.ImageCollection
        :param kwargs:
        :return:
        """

        reduxFirstImg = ImageEE(
            coll_ee.filter(
                self.filters
            ).first()
        )
        return reduxFirstImg

@register(factory)
@register_all(__all__)
class PHEStartSeason(PHESeason):
    """ Start growing season filter

    :param stage: starting time placeholder 's'
    :type stage: str

    """
    def __init__(self, **kwargs):

        if kwargs.has_key("stage"):
            if not kwargs["stage"] == "s":
                kwargs["stage"] = "s"
        else:
            kwargs["stage"] = "s"

        super(PHEStartSeason, self).__init__(**kwargs)

@register(factory)
@register_all(__all__)
class PHEEndSeason(PHESeason):
    """ End growing season filter

    :param stage: ending time placeholder 'e'
    :type stage: str

    """
    def __init__(self, **kwargs):

        if kwargs.has_key("stage"):
            if not kwargs["stage"] == "e":
                kwargs["stage"] = "e"
        else:
            kwargs["stage"] = "s"

        super(PHEEndSeason, self).__init__(**kwargs)

@register(factory)
@register_all(__all__)
class GrowingSeasonDate(PHESeason):
    """ Growing season start/end can span over three years

    """
    def __init__(self, **kwargs):

        if 'year' in kwargs.keys():
            self.year = kwargs.get("year")
        elif self.filters:
            for filter in self.filters:
                if 'time_extent' in filter["leftField"]:
                    self.year = filter["rightValue"][5:9]
        else:
            pass
        if not self.year:
            raise GEEException("Year is not defined")

        super(GrowingSeasonDate, self).__init__(**kwargs)

    def apply(self):
        """ Apply the filter

        :param year: define filter dates around target year. Goes from 1970
        to now. Growing season start/end can span over three years
        :type year: str
        :param kwargs:
        :return:
        """
        self.logger.debug("The target year for growing season is {0}".format(
            self.year
        ))
        GEE_dates = (
            DateEE.fromYMD(int(self.year) - 1, 1, 1),
            DateEE.fromYMD(int(self.year) + 1, 12, 31)
        )
        return GEE_dates



# -*- coding: utf-8 -*-
""" Registry Decorator """

import functools

__all__ = ('register', 'register_all')

def register(holder):
    """ Make a registry of the decorated Class in the given holder

    :Usage:

    .. code:: python

        from registry_decorator import register
        from gee_pheno import Phenology

        registry = {}

        @register(registry)
        class NewPhenology(Phenology):
            def __init__(**kwargs):
                pass

    :param holder: dict that will hold the classes
    :type holder: dict
    """
    def wrap(cls):
        name = cls.__name__
        holder[name] = cls
        return cls
    return wrap

def register_all(holder):
    """ Make a registry of the decorated Class in the given holder to use in
    module's __all__ variable

    :Usage:

    .. code:: python

        from registry_decorator import register
        from gee_pheno import Phenology

        __all__ = []

        @register_all(__all__)
        class NewPhenology(Phenology):
            def __init__(**kwargs):
                pass

    :param holder: list that will hold the classe's names
    :type holder: list
    """
    def wrap(cls):
        name = cls.__name__
        holder.append(name)
        return cls
    return wrap


# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Vienna University of Technology - Department of
#   Geodesy and Geoinformation nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Thomas Mistelbauer thomas.mistelbauer@geo.tuwien.ac.at
# Creation date: 2014-08-05

"""
This module provides functions for date manipulation on a dekadal basis.

A dekad is defined as days 1-10, 11-20 and 21-last day of a month.

Or in numbered dekads:

1: day 1-10
2: day 11-20
3: day 21-last
"""

import calendar
import pandas as pd
import math
from datetime import datetime


def dekad_index(begin, end=None):
    """Creates a pandas datetime index on a decadal basis.

    Parameters
    ----------
    begin : datetime
        Datetime index start date.
    end : datetime, optional
        Datetime index end date, set to current date if None.

    Returns
    -------
    dtindex : pandas.DatetimeIndex
        Dekadal datetime index.
    """

    if end is None:
        end = datetime.now()

    mon_begin = datetime(begin.year, begin.month, 1)
    mon_end = datetime(end.year, end.month, 1)

    daterange = pd.date_range(mon_begin, mon_end, freq='MS')

    dates = []

    for i, dat in enumerate(daterange):
        lday = calendar.monthrange(dat.year, dat.month)[1]
        if i == 0 and begin.day > 1:
            if begin.day < 11:
                if daterange.size == 1:
                    if end.day < 11:
                        dekads = [10]
                    elif end.day >= 11 and end.day < 21:
                        dekads = [10, 20]
                    else:
                        dekads = [10, 20, lday]
                else:
                    dekads = [10, 20, lday]
            elif begin.day >= 11 and begin.day < 21:
                if daterange.size == 1:
                    if end.day < 21:
                        dekads = [20]
                    else:
                        dekads = [20, lday]
                else:
                    dekads = [20, lday]
            else:
                dekads = [lday]
        elif i == (len(daterange) - 1) and end.day < 21:
            if end.day < 11:
                dekads = [10]
            else:
                dekads = [10, 20]
        else:
            dekads = [10, 20, lday]

        for j in dekads:
            dates.append(pd.datetime(dat.year, dat.month, j))

    dtindex = pd.DatetimeIndex(dates)

    return dtindex


def check_dekad(date):
    """Checks the dekad of a date and returns the dekad date.

    Parameters
    ----------
    date : datetime
        Date to check.

    Returns
    -------
    new_date : datetime
        Date of the dekad.
    """
    if date.day < 11:
        dekad = 10
    elif date.day > 10 and date.day < 21:
        dekad = 20
    else:
        dekad = calendar.monthrange(date.year, date.month)[1]

    new_date = datetime(date.year, date.month, dekad)

    return new_date


def dekad_startdate_from_date(dt_in):
    """
    dekadal startdate that a date falls in

    Parameters
    ----------
    run_dt: datetime.datetime

    Returns
    -------
    startdate: datetime.datetime
        startdate of dekad
    """
    if dt_in.day <= 10:
        startdate = datetime(
            dt_in.year,
            dt_in.month,
            1, 0, 0, 0
        )
    if dt_in.day >= 11 and dt_in.day <= 20:
        startdate = datetime(
            dt_in.year,
            dt_in.month,
            11, 0, 0, 0
        )
    if dt_in.day >= 21:
        startdate = datetime(
            dt_in.year,
            dt_in.month,
            21, 0, 0, 0
        )
    return startdate


def check_dekad_enddate(dt):
    """
    Check if a date is a dekad enddate
    """
    return check_dekad(dt) == dt


def check_dekad_startdate(dt):
    """
    Check if a date is a dekad startdate
    """
    if dt.day in [1, 11, 21]:
        return True
    else:
        return False


def group_into_dekads(dates, use_dekad_startdate=False):
    """
    Group a list of dates into dekads.

    Parameters
    ----------
    dates: list of datetime.datetime
    use_dekad_startdates: boolean, optional
        If set the dekad reference dates will
        be the startdates of the dekad

    Returns
    -------
    groups: dict
        keys: dekad reference dates
        values: list of dates belonging to dekad
    """
    groups = {}
    for dt in dates:
        dekad_date = check_dekad(dt)
        if use_dekad_startdate:
            dekad_date = dekad_startdate_from_date(dekad_date)
        if dekad_date not in groups:
            groups[dekad_date] = []
        groups[dekad_date].append(dt)
    return groups


def dekad2day(year, month, dekad):
    """Gets the day of a dekad.

    Parameters
    ----------
    year : int
        Year of the date.
    month : int
        Month of the date.
    dekad : int
        Dekad of the date.

    Returns
    -------
    day : int
        Day value for the dekad.
    """

    if dekad == 1:
        day = 10
    elif dekad == 2:
        day = 20
    elif dekad == 3:
        day = calendar.monthrange(year, month)[1]

    return day


def runningdekad2date(year, rdekad):
    """Gets the date of the running dekad of a specifc year.

    Parameters
    ----------
    year : int
        Year of the date.
    rdekad : int
        Running dekad of the date.

    Returns
    -------
    datetime.datetime
        Date value for the running dekad.
    """

    month = int(math.ceil(rdekad / 3.))
    dekad = rdekad - month * 3 + 3
    day = dekad2day(year, month, dekad)

    return datetime(year, month, day)


def day2dekad(day):
    """Returns the dekad of a day.

    Parameters
    ----------
    day : int
        Day of the date.

    Returns
    -------
    dekad : int
        Number of the dekad in a month.
    """

    if day < 11:
        dekad = 1
    elif day > 10 and day < 21:
        dekad = 2
    else:
        dekad = 3

    return dekad


def get_dekad_period(dates):
    """Checks number of the dekad in the current year for dates given as list.

    Parameters
    ----------
    dates : list of datetime
        Dates to check.

    Returns
    -------
    period : list of int
        List of dekad periods.
    """

    period = []

    for dat in dates:

        d = check_dekad(dat)
        dekad = day2dekad(d.day)
        per = dekad + ((d.month - 1) * 3)
        period.append(per)

    return period



""" Module holding custom conversion from dedaks to millis for images """

from ee import Image, EEException, Date, List
import daiquiri
from exception import GEEException

class PHEDekad2Millis(object):
    def __init__(self, phe_img=None, tyear=None, **kwargs):

        logger = daiquiri.getLogger(__name__, subsystem="dekad2millis")
        self.logger = logger

        if not phe_img and not isinstance(phe_img, Image):
            raise TypeError("Phenology image is not correct")
        else:
            self.phe_img = phe_img

        self.logger.info("Target Year will be calculated from image")
        try:
            self.tyear = Date(
                phe_img.get('system:time_start')
            ).get('year')
        except EEException as e:
            raise GEEException(e.message)

        super(PHEDekad2Millis, self).__init__()

    @property
    def phey(self):  # @TODO check if dekad.py can help more
        """ Assign Phenology year based on dekad number
            0-36 = Target year - 1
            37-72 = Target year
            73 - 108 = Target year + 1, else = 0
        """

        return self.phe_img.expression(
            "(arg >108) ? 0 \
            : (arg >72) ? Ty+1 \
            : (arg >36) ? Ty \
            : Ty-1", {
                'arg': self.phe_img.select(0),
                'Ty': self.tyear
            }
        )

    @property
    def phem(self):  # @TODO check if dekad.py can help more
        """ Assign Phenology month based on dekad number
            0-36 = dekad/3 (rounded up)
            37-72 = (dekad-36)/3 (rounded up)
            73 - 108 = (dekad-72)/3 (rounded up)
        """

        return self.phe_img.expression(
            "(arg >108) ? 0 \
            : (arg >72) ? int(round((arg-72)/3+0.2)) \
            : (arg >36) ? int(round((arg-36)/3+0.2)) \
            : int(round((arg/3)+0.2))", {
                'arg': self.phe_img.select(0)
            }
        )

    @property
    def phed(self):  # @TODO check if dekad.py can help more
        """ Assign Phenology day based on dekad number: dekad/3-int(dekad/3)
            if > 0.6 then day = 11 (8/3 = 2.66 = second dekad  = 11)
            if > 0.3 then day = 1 (7/3 = 2.33 = first dekad  = 1)
            if < 0.3 then day = 21 (9/3 = 3.0 = third dekad  = 21)
        """

        return self.phe_img.expression(
            "(arg/3-int(arg/3))>0.6 ? 11 \
            :(arg/3-int(arg/3))>0.3 ? 1 \
            : 21", {
                'arg': self.phe_img.select(0)
            }
        )

    def _phe_bands_ymd_generator(self):
        """ Add generated YMD bands to phenology image

        """

        return self.phe_img.cat(
            [self.phey, self.phem.byte(), self.phed]
        ).rename(['PHEsY', 'PHEsM', 'PHEsD'])

    def _startyear2millis(self, y):
        #  Use pendulum?
        datey = "{0}-01-01".format(
            y
        )
        return Date(datey).millis()

    def _set_ymd_out(self, startY, endY):
        ee_ymillis_list = []
        for year in range(startY, endY + 1):
            ee_ymillis_list.append(
                self._startyear2millis(year).getInfo()
            )
        ee_mmillis_list = [
            0,
            2678400000,
            5097600000,
            7776000000,
            10368000000,
            13046400000,
            15638400000,
            18316800000,
            20995200000,
            23587200000,
            26265600000,
            28857600000
        ]
        ee_dmillis_list = [0, 864000000, 1728000000]
        return ee_ymillis_list, ee_mmillis_list, ee_dmillis_list

    def _datemillis(self, syear, eyear):
        dekads = [1, 11, 21]

        # yearly millis
        ymillis = self._phe_bands_ymd_generator().select(
            "PHEsY"
        ).interpolate(
            List.sequence(
                syear - 1,
                eyear + 1,
                1
            ),
            self._set_ymd_out(syear - 1, eyear + 1)[0],
            'mask'
        )
        self.logger.debug(
            "PHEsY interpolation image result {0}".format(
                ymillis.getInfo()
            )
        )
        # monthly millis
        mmillis = self._phe_bands_ymd_generator().select(
            "PHEsM"
        ).interpolate(
            List.sequence(
                1,
                12,
                1
            ),
            self._set_ymd_out(syear - 1, eyear + 1)[1],
            'mask'
        )
        self.logger.debug(
            "PHEsM interpolation image result {0}".format(
                mmillis.getInfo()
            )
        )
        # daily millis
        dmillis = self._phe_bands_ymd_generator().select(
            "PHEsD"
        ).interpolate(
            dekads,
            self._set_ymd_out(syear - 1, eyear + 1)[2],
            'mask'
        )
        self.logger.debug(
            "PHEsD interpolation image result {0}".format(
                dmillis.getInfo()
            )
        )

        return self.phe_img.cat(
            self._phe_bands_ymd_generator(),
            ymillis.add(mmillis).add(dmillis)
        )

    def apply(self, yos, yoe):
        return self.phe_img.addBands(
            self._datemillis(yos, yoe)
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

from glob import glob

from os.path import (
    basename,
    splitext
)

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'earthengine-api',
    'daiquiri',
    'pandas'
]

setup_requirements = [
    'pytest-runner',
    'earthengine-api',
    'daiquiri',
    'pandas'
]

test_requirements = ['pytest', ]
setup(
    author="Francesco Bartoli",
    author_email='francesco.bartoli@geobeyond.it',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="GEE phenology",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='gee_pheno',
    name='gee_pheno',
    packages=find_packages(),
    py_modules=[splitext(basename(path))[0] for path in glob('gee_pheno/*.py')],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/francbartoli/gee_pheno',
    version='0.1.5',
    zip_safe=False,
)
